

// %CopyrightBegin%
// 
// Copyright Ericsson AB 1997-2016. All Rights Reserved.
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// %CopyrightEnd%


enum E1 {kalle, sune};

enum E2 { el0, el1, el2, el3, el4, el5, el6, el7, el8, el9, el10, el11, el12, el13,
el14, el15, el16, el17, el18, el19, el20, el21, el22, el23, el24, el25, el26, el27,
el28, el29, el30, el31, el32, el33, el34, el35, el36, el37, el38, el39, el40, el41,
el42, el43, el44, el45, el46, el47, el48, el49, el50, el51, el52, el53, el54, el55,
el56, el57, el58, el59};





