

// %CopyrightBegin%
// 
// Copyright Ericsson AB 1998-2016. All Rights Reserved.
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// %CopyrightEnd%
#define T 12
#define F T

//comment
/*exception except {};*/

// comment
  // comment
/* another */
   /* another */
/* still
another */
  /* still
  another */
__LINE__
/* yet \
  another */
// yet \
  another
__LINE__

#include "all.c"
#include <all.c>
#include /* comment */ "all.c"
#include /* comment */ <all.c>
#include  "all.c" /* comment */
#include  <all.c> /* comment */
#include // "all.c"
#include // <all.c>
#include  "all.c" // comment
#include  <all.c> // comment
#include "all/*cc*/.c"
#include <all/*cc*/.c>

main()
{
  printf(" %d \n",F);
  a();

}
//comment
/*exception hell {};*/
#undef T 
#define T "3/*com\
ment*/4"
a()
{
  printf(" %d \n",F);
  printf(" %d \n",T);
}

b()
{}

