

// %CopyrightBegin%
// 
// Copyright Ericsson AB 1997-2016. All Rights Reserved.
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// %CopyrightEnd%


// Check multiply defined declarators

enum E2 {kal, kula, E1};	// legal, but used below

// Now check that declarator a is multiply defined in all unions below
union U0 switch (long) {
case 0: long a;
case 1: short a;
};
union U00 switch (char) {
case 'c' : long a;
case 'f' : char c;
case 'b' : short a;
};
union U000 switch (boolean) {
case TRUE: long a;
case FALSE: short a;
};
union U0000 switch (E2) {
case kal: long a;
case kula: short a;
};




// Check that enum name duplication is found.

union U1 switch (enum E1 {kalle, kula, E1}) {
case E1	: long a;	// legal
case kalle	: short E1;	// illegal
};


// This is legal, but ended up here anyway

union U2 switch(::E2) {
case kal	: long a;
case kula	: short b;
default		: boolean E1;
};
