

// %CopyrightBegin%
// 
// Copyright Ericsson AB 1998-2016. All Rights Reserved.
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// %CopyrightEnd%
#line
#line 8
#line 8a
#line 12 abc.c
#line 12 "kurt.c"
#warning fdafdsaf 


#define T 12
#define F T
#define Q(a) sinus(a)
#undef Q
#
#line 12
#warning test of warning
#warning second of warning
#warning third of warning
#pragma kurt
#ident kurt
#kurt      fdsafd
#line 20
main()
{
  if (Q(34,56)=='NULL') printf(" T AAA%sEEEE \n",Q);
  printf(" %d \n",F);
}
sune
