

// %CopyrightBegin%
// 
// Copyright Ericsson AB 1997-2016. All Rights Reserved.
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// %CopyrightEnd%

// Check that case labels are not duplicated

union U1 switch (long) {
case 1 : long a;
case 1 : short b;
};

union U2 switch (char) {
case 'c' : long a;
case 'c' : short b;
};

union U2b switch (char) {
case 'c' : 
case 'c' : long a;
case 'e': long b;
case 'c': long c;
};

union U3 switch (enum E1 {kalle, kula}) {
case kula : long a;
case kula : short b;
};

union U4 switch (boolean) {
case TRUE : long a;
case TRUE : short b;
};

union U5 switch (boolean) {
case TRUE : long a;
default: short p;
default: short pp;
};

