/*
 * %CopyrightBegin%
 * 
 * Copyright Ericsson AB 1998-2016. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * %CopyrightEnd%
 */
#ifndef _ERL_GLOBAL_H
#define _ERL_GLOBAL_H

char **erl_global_names(int fd, int *count);
ETERM *erl_global_whereis(int fd, const char *name, char *node);
int erl_global_register(int fd, const char *name, ETERM *pid);
int erl_global_unregister(int fd, const char *name);

#endif /* _ERL_GLOBAL_H */
