

// %CopyrightBegin%
// 
// Copyright Ericsson AB 1998-2016. All Rights Reserved.
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// %CopyrightEnd%
//
// IDL for testing register/unregister in the IFR when using included specs
// Special case with multiple inheritence.
//
#include "reg_m10.idl"
#include "reg_m11.idl"

module m4 {

  interface i4 : m2::i2
    {
      short op5( in long a, inout char b, out long c );
    };

  interface i5 : m3::i3
    {
      short op6( in long a, inout char b, out long c );
    };


};

