/*
 * %CopyrightBegin%
 * 
 * Copyright Ericsson AB 1999-2016. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * %CopyrightEnd%
 */

#include "zlib.h"

typedef struct erts_gzFile* ErtsGzFile;

ErtsGzFile erts_gzopen (const char *path, const char *mode);
int erts_gzread(ErtsGzFile file, voidp buf, unsigned len);
int erts_gzwrite(ErtsGzFile file, voidp buf, unsigned len);
int erts_gzseek(ErtsGzFile, int, int);
int erts_gzflush(ErtsGzFile file, int flush);
int erts_gzclose(ErtsGzFile file);
ErlDrvBinary* erts_gzinflate_buffer(char*, uLong);
ErlDrvBinary* erts_gzdeflate_buffer(char*, uLong);
