

// %CopyrightBegin%
// 
// Copyright Ericsson AB 1998-2016. All Rights Reserved.
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// %CopyrightEnd%
#11a
#define xstr str(s) + kurt*2;
#define asdf pragma
#asdf
#define asd #pragma asd

#10
#12 8kurt

#define sss "stringing in the rain"
#define ddd "string
ing in the rain" asd
#line 20
#include "head.h" qqqq
#include %!#
#include <sys.h>

