%%------------------------------------------------------------------------
%%
%% %CopyrightBegin%
%% 
%% Copyright Ericsson AB 2001-2016. All Rights Reserved.
%% 
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.
%% 
%% %CopyrightEnd%
%%
%%
%%------------------------------------------------------------------------
%% Description: a very simple implementation of Push Supplier interface
%%------------------------------------------------------------------------
 
-module(event_test_PushS_impl).
 
-export([init/1, terminate/2, disconnect_push_supplier/1, do_push/2]).
 
init(Proxy) ->
        {ok, Proxy}.
 
terminate(_From, _Reason) ->
    ok.

disconnect_push_supplier(Proxy) ->
    io:format("event_test_PullC terminates~n",[]),
    {stop, normal, ok, Proxy}.

do_push(Proxy, Event) ->
    {reply, 'CosEventComm_PushConsumer':push(Proxy, Event), Proxy}.

