%%
%% %CopyrightBegin%
%% 
%% Copyright Ericsson AB 1999-2016. All Rights Reserved.
%% 
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.
%% 
%% %CopyrightEnd%
%%
%%
%%------------------------------------------------------------
%%
%% Example 
%%
%%------------------------------------------------------------
-module(m_i_impl).

%% Standard functions
-export([init/1, terminate/2]).
%% Interface functions
-export([f/2, g/2]).

init(_Env) ->
    {ok, []}.

terminate(_From, _Reason) ->
    ok.

f(State, In) ->
    io:format("f working ....\n", []),
    {reply, In, State}.

g(State, _In)  ->
    io:format("g working ....\n", []),
    {noreply, State}.





