

// %CopyrightBegin%
// 
// Copyright Ericsson AB 1998-2016. All Rights Reserved.
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// %CopyrightEnd%
// Bad pragma IDs
 
// Completelly bad id
module M1 {
  
  typedef long T1;
 
  typedef long T2;
 
#pragma ID T2 "CompletelyBadId"
 
};


// Bad id, should start with DCE
module M2 {
  
  typedef long T1;
 
  typedef long T2;
 
#pragma ID T2 "BAD:d62207a2-011e-11ce-88b4-0800090b5d3e:3"
 
};
 

// Bad version in ID : not a short number
module M3 {
  
  typedef long T1;
 
  typedef long T2;
 
#pragma ID T2 "DCE:d62207a2-011e-11ce-88b4-0800090b5d3e:ABCD"
 
};

 
 


















