//
// %CopyrightBegin%
//
// Copyright Ericsson AB 1999-2016. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// %CopyrightEnd%
//

#ifndef _TEST_IDL
#define _TEST_IDL
#include <../src/CosTransactions.idl>
module etrap_test { 

  // interface server
  interface server :
    CosTransactions::Coordinator, CosTransactions::SubtransactionAwareResource, 
    CosTransactions::RecoveryCoordinator, CosTransactions::Control {
  };
//   interface Server :
//     CosTransactions::Coordinator, CosTransactions::SubtransactionAwareResource, 
//     CosTransactions::RecoveryCoordinator, CosTransactions::Control,
//     CosTransactions::Synchronization {
//   };

}; // End of test Module

#endif
