

// %CopyrightBegin%
// 
// Copyright Ericsson AB 1997-2016. All Rights Reserved.
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// %CopyrightEnd%

//
// Checking that default labels are correct in TK
//

interface i1 {
  union U1 switch (long) {
  default: long a;
  case 1: case 2: long b;
  };

  union U2 switch (long) {
  case 0: default: long a;
  case 1: case 2: long b;
  };

  union U3 switch (long) {
  case -1: long aa;
  case 0: default: long a;
  case 1: case 2: long b;
  };
  
  union U4 switch (long) {
  case -1: long aa;
  case 0: long a;
  case 1: case 2: long b;
  };
  
  U1 op0();
  U2 op1();
  U3 op2();
  U4 op3();
};
